(* Copyright 2008, 2009 Michael Brunner *)

(* This file is part of LIPS.

  LIPS is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as 
	published by the Free Software Foundation, either version 3 of 
	the License, or (at your option) any later version.

  LIPS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
	License along with LIPS.  If not, see 
	<http://www.gnu.org/licenses/>. *)

type var = Strng.t * int;;
type cons = True | False;;
type t = Var of var | Cons of cons | Abs of var * t | App of t * t | Let of var * t * t | If of t * t * t;;
type dir = Left | Middle | Right;;
type pos = dir list;;


let var x = Var (x, 0);;
let abs xs body = Lst.fold (fun x xs -> Abs ((x, 0), xs)) body xs;;
let app ts = Lst.fold_left1 (fun s t -> App (s, t)) ts;; 
let leti v t body = Let ((v, 0), t, body);;
let ite c i e = If (c, i, e);;
let cons c = Cons c;;



(** retrieve all variables *)
let rec vars = function
	| Var x -> St.singleton x
	| Abs (x, t) -> St.insert x (vars t)
	| App (t1, t2) -> St.union (vars t1) (vars t2)
	| Let (v, t, b) -> St.insert v (St.union (vars t) (vars b))
	| If (c, t1, t2) -> St.union (St.union (vars c) (vars t1)) (vars t2)
	| Cons _ -> St.empty;;

(** retrieve all free variables *)
let rec fvars = function
	| Var x -> St.singleton x
	| Abs (v, t) -> St.diff (fvars t) (St.singleton v)
	| App (t1, t2) -> St.union (fvars t1) (fvars t2)
	| Let (v, t, b) -> St.union (St.diff (fvars b) (St.singleton v)) (fvars t)
	| If (c, t1, t2) -> St.union (St.union (fvars c) (fvars t1)) (fvars t2)
	| Cons _ -> St.empty;;


let max_index v t = 
	let check_var (v',i) = if v' = v then i else -1 in 
	let rec max_index2 = function 
		| Var x -> check_var x
		| Abs (v, t) -> max (check_var v) (max_index2 t)
		| App (t1, t2) -> max (max_index2 t1) (max_index2 t2) 
		| Let (v, t, b) -> max (max (check_var v) (max_index2 t)) (max_index2 b)
		| If (c, t1, t2) -> max (max (max_index2 c) (max_index2 t1)) (max_index2 t2)
		| Cons _ -> -1
	in max_index2 t;;


(** substitute term for variable *)
let rec substitute x t = function
	| Var y as v -> if y = x then t else v
	| Abs (v, u) as s -> if v = x then s else if St.mem v (fvars t) then let v' = (fst v, (max_index (fst v) t) + 1) in Abs (v', substitute x t (substitute v (Var v') u)) else Abs (v, substitute x t u)
	| App (t1, t2) -> App (substitute x t t1, substitute x t t2)
	| Let (v, d, b) -> if v = x then Let (v, substitute x t d, b) else if St.mem v (fvars t) then let v' = (fst v, (max_index (fst v) t) + 1) in Let (v', substitute x t d, substitute x t (substitute v (Var v') b)) else Let (v, substitute x t d, substitute x t b)
	| If (c, t1, t2) -> If (substitute x t c, substitute x t t1, substitute x t t2)
	| cons -> cons;;

(** beta-step *)
let beta = function
	| App (Abs (x, u), t) -> substitute x t u
	| Let (v, t, b) -> substitute v t b
	| If (Cons True, t1, _) -> t1
	| If (Cons False, _ , t2) -> t2
	| _ -> failwith "CoreML.beta: term is no redex";;


(** check if term is redex *)
let is_redex = function 
	| App (Abs _, _) -> true
	| Let (_, _, _) -> true
	(* if only counts as redex if the condition is evaluated *)
	| If (Cons True, _, _) -> true
	| If (Cons False, _, _) -> true
	| _ -> false;;

(** retrieve all redex-positions *)
let redexpos term = 
	let rec redexes p r = function 
		| Var _ -> r
		| Abs (_, t) -> redexes (Left::p) r t
		(*| App (Abs _, t) -> redexes (Left::p) ((Lst.rev p)::r) t*)
		| App (Abs (_, s), t) -> (Lst.rev (p))::(redexes (Left::(Left::p)) r s) @ (redexes (Right::p) r t)
		| App (t1, t2) -> (redexes (Left::p) r t1) @ (redexes (Right::p) r t2)
		| Let (v, t, b) -> (Lst.rev p)::(redexes (Right::p) r b)
		| If (Cons True, _, _) -> (Lst.rev p)::r
		| If (Cons False, _, _) -> (Lst.rev p)::r
		| If (c, t1, t2) -> (redexes (Left::p) r c) (* only report the redexes of condition and thereby force outer-evaluation of if-statements *)
		| Cons _ -> r
	in redexes [] [] term;; 

(** check wether term is in weak-head normalform *)
let rec is_whnf t = 
	Lst.mem [] (redexpos t) || match t with If (c, _, _) -> is_whnf c | _ -> false;;

(** check wether term is normalform *)
let is_nf t = (redexpos t = []);;

(** check wether term is closed or not *)
let is_closed t = St.is_empty (fvars t);;



(** retrieve the positions of a subterm within term *)
let pos sterm term = 
	let rec pos p r = function 
		| Var _ as x -> if x = sterm then (Lst.rev p)::r else r
		| Abs (v, t) as x -> if x = sterm then (Lst.rev p)::r else pos (Left::p) r t
		| App (t1, t2) as x -> if x = sterm then (Lst.rev p)::r else (pos (Left::p) r t1) @ (pos (Right::p) r t2)
		| Let (v, t, b) as x -> if x = sterm then (Lst.rev p)::r else (pos (Left::p) r t) @ (pos (Right::p) r b)
		| If (c, t1, t2) as x -> if x = sterm then (Lst.rev p)::r else (pos (Left::p) r c) @ (pos (Middle::p) r t1) @ (pos (Right::p) r t2)
		| Cons _ as x -> if x = sterm then (Lst.rev p)::r else r
	in pos [] [] term;;

(** retrieve subterm according to position *)
let rec get term = function 
	| [] -> term
	| i::ps -> match term with 
		| Var _ -> failwith "CoreML.get: position does not match term"
		| Abs (v, t) -> get t ps 
		| App (t1, t2) -> get (if i = Left then t1 else t2) ps
		| Let (v, t, b) -> get (if i = Left then t else b) ps
		| If (c, t1, t2) -> get (match i with Left -> c | Middle -> t1 | Right -> t2) ps
		| Cons _ -> failwith "CoreML.get: position does not match term";;

(** replace subterm at given position with new term *)
let replace term newterm pos= 
	let rec replace term = function 
	| [] -> newterm
	| i::ps -> match term with 
		| Var _ -> failwith "CoreML.replace: position does not match term"
		| Abs (v, t) -> Abs (v, replace t ps) 
		| App (t1, t2) -> if i = Left then App (replace t1 ps, t2) else App (t1, replace t2 ps) 
		| Let (v, t, b) -> if i = Left then Let (v, replace t ps, b) else Let (v, t, replace b ps)
		| If (c, t1, t2) -> begin match i with 
			| Left -> If (replace c ps, t1, t2)
			| Middle -> If (c, replace t1 ps, t2)
			| Right -> If (c, t1, replace t2 ps) end 
		| Cons _ -> failwith "CoreML.replace: position does not match term"	
	in replace term pos;;

let rec replace_all term = function 
	| [] -> term
	| (p, t)::subst -> replace_all (replace term t p) subst;;

(** replace all occurencies of variable with new one *)
let rec replace_var (m, j) r = function 
	| Var (n, i) as t -> if (n = m) && (i = j) then Var r else t
	| Abs ((n, i), t) -> if (n = m) && (i = j) then Abs (r, replace_var(m, j) r t) else Abs ((n, i), replace_var (m, j) r t) 
	| App (t1, t2) -> App (replace_var (m, j) r t1, replace_var (m, j) r t2) 
	| Let ((n, i), t, b) -> if (n = m) && (i = j) then Let (r, replace_var (m, j) r t, replace_var (m, j) r b) else Let ((n, i), replace_var (m, j) r t, replace_var (m, j) r b)
	| If (c, t1, t2) -> If (replace_var (m, j) r c, replace_var (m, j) r t1, replace_var (m, j) r t2)
	| cons -> cons;; 

(** replace all variables according to a given substitution *)
let rec replace_all_vars term = function 
	| [] -> term 
	| ((n1,i1),(n2,i2))::xs -> replace_all_vars (replace_var (n1, i1) (n2, i2) term) xs;;

(** reduce the given term's variable indexs to minimum *)
let reduce_vars term = 
	let next_var_index var_name index_list = Lst.fold (fun (n, i) xs -> (if n = var_name then i else 0) + xs) 0 index_list in 
	let inc_var_index var_name index_list = Lst.map (fun (n, i) -> if n = var_name then (n, i - 1) else (n, i)) index_list in
	let rec get_substitution index_list = function  
		| [] -> []
		| (n, i)::vs -> ((n, i), (n, next_var_index n index_list))::(get_substitution (inc_var_index n index_list) vs) in
	let var_list = (St.to_list (vars term)) in 
	let index_list = St.to_list (Lst.fold (fun x xs -> St.insert (x, 0) xs) St.empty (Lst.map (fun (n, i) -> n) var_list)) in
	let subst = get_substitution index_list var_list in
	let resubst = Lst.map (fun ((n1, i1), (n2,i2)) -> ((n2, i2), (n2, -i2))) subst in
	replace_all_vars (replace_all_vars term subst) resubst;;




module Strategies = struct 
	
	(* call by value: innermost but don't reduce whnf *)
	let call_by_value term = 
		let rec comp p1 p2 = match (p1, p2) with
			| ([], []) -> 0
			| ([], _) -> 1
			| (_, []) -> -1
			| (Left::_, Right::_) -> -1
			| (Left::_, Middle::_) -> -1
			| (Middle::_, Right::_) -> -1
			| (Middle::_, Left::_) -> 1
			| (Right::_, Left::_) -> 1
			| (Right::_, Middle::_) -> 1
			| (_::xs, _::ys) -> comp xs ys in
		let next_beta term = let rlist = (*Lst.filter (fun x -> not (is_whnf (get term x)))*) (Lst.sort comp (redexpos term)) in match rlist with
			| [] -> []
			| p:: _ -> [p] in
		let rec get_subst term = function
			| [] -> []
			| p:: ps -> (p, beta (get term p)):: (get_subst term ps)
		in get_subst term (next_beta term);;
		
	(* call by name: outermose but don't reduce whnf *)
  let call_by_name term = 
		let rec comp p1 p2 = match (p1, p2) with
			| ([], []) -> 0
			| ([], _) -> -1
			| (_, []) -> 1
			| (Left::_, Right::_) -> -1
			| (Left::_, Middle::_) -> -1
			| (Middle::_, Right::_) -> -1
			| (Middle::_, Left::_) -> 1
			| (Right::_, Left::_) -> 1
			| (Right::_, Middle::_) -> 1
			| (_:: xs, _:: ys) -> comp xs ys in
		let next_beta term = let rlist = (*Lst.filter (fun x -> not (is_whnf (get term x)))*) (Lst.sort comp (redexpos term)) in match rlist with
			| [] -> []
			| p:: _ -> [p] in
		let rec get_subst term = function
			| [] -> []
			| p:: ps -> (p, beta (get term p)):: (get_subst term ps)
	in get_subst term (next_beta term);;

  let parallel_inner term = 
		let rec comp p1 p2 = match (p1, p2) with
			| ([], []) -> 0
			| ([], _) -> 1
			| (_, []) -> -1
			| (Left::_, Right::_) -> -1
			| (Left::_, Middle::_) -> -1
			| (Middle::_, Right::_) -> -1
			| (Middle::_, Left::_) -> 1
			| (Right::_, Left::_) -> 1
			| (Right::_, Middle::_) -> 1
			| (_::xs, _::ys) -> comp xs ys in
		let rec is_parallel p1 p2 = match (p1, p2) with
			| ([], []) -> false
			| (_, []) -> false
			| ([], _) -> false
			| (d1:: _, d2:: _) when d1 != d2 -> true
			| (_:: xs, _:: ys) -> is_parallel xs ys in
		let rec get_parallel ps cs = match cs with 
			| [] -> ps
			| c::cs -> if (Lst.fold1 (fun x y -> x && y) (Lst.map (is_parallel c) ps)) then get_parallel (c::ps) cs else get_parallel ps cs in
		let next_beta term =
			let redp = Lst.sort comp (redexpos term)
			in match redp with
				| [] -> []
				| p:: ps -> get_parallel [p] ps in
		let rec get_subst term = function
			| [] -> []
			| p:: ps -> (p, beta (get term p)):: (get_subst term ps)
	in get_subst term (next_beta term);;

  let parallel_outer term = 
		let rec comp p1 p2 = match (p1, p2) with
			| ([], []) -> 0
			| ([], _) -> -1
			| (_, []) -> 1
			| (Left::_, Right::_) -> -1
			| (Left::_, Middle::_) -> -1
			| (Middle::_, Right::_) -> -1
			| (Middle::_, Left::_) -> 1
			| (Right::_, Left::_) -> 1
			| (Right::_, Middle::_) -> 1
			| (_:: xs, _:: ys) -> comp xs ys in
		let rec is_parallel p1 p2 = match (p1, p2) with
			| ([], []) -> false
			| (_, []) -> false
			| ([], _) -> false
			| (d1:: _, d2:: _) when d1 != d2 -> true
			| (_:: xs, _:: ys) -> is_parallel xs ys in
		let rec get_parallel ps cs = match cs with 
			| [] -> ps
			| c::cs -> if (Lst.fold1 (fun x y -> x && y) (Lst.map (is_parallel c) ps)) then get_parallel (c::ps) cs else get_parallel ps cs in
		let next_beta term =
			let redp = Lst.sort comp (redexpos term)
			in match redp with
				| [] -> []
				| p:: ps -> get_parallel [p] ps in
		let rec get_subst term = function
			| [] -> []
			| p::ps -> (p, beta (get term p)):: (get_subst term ps)
	in get_subst term (next_beta term);;

end;;

let do_step strat term = replace_all term (strat term);;
let rec calc_steps strat term = 
	let ps = Lst.map (fun (p, _) -> p) (strat term) in
	if (*is_whnf term ||*) (*is_nf term*) ps = [] then 
		[(term, [])] 
	else 
		(term, ps) :: (calc_steps strat (do_step strat term));;
let calc_nf strat term = 
	let (result, _) = Lst.last (calc_steps strat term) in result;;



(* PARSERS *)
open Parser;; (* needed for infix-operators of module Parser *)
(* simple parser for parenthesis, this one only checks wether
   each '(' is followed by one ')', hence more ')' than needed 
	 may occur *)
let parenthesis p = char '(' >> p >>= fun result -> char ')' >> return result;;

(* parser for variable-identifiers *)
let varname term = 
	((many1 (char '_' <|> letter <|> digit)) >>= fun p1 -> (many (char '\'')) >>= fun p2 -> return (p1 @ p2)) term;;

(** fully functional CoreML-parser, with arbitrary long variable-names **)
let rec ltp term = (letp <|> ifp <|> consp <|> appp <|> absp <|> varp) term
  and pltp term = (parenthesis ltp) term
	and absp term = (char '\\' >> abshp >>= fun t1 -> (char '.' >> spaces) >> absbp >>= fun t2 -> return (abs t1 t2)) term
	and abshp = sep_end_by1 (varname) spaces1
	and absbp term = (ltp <|> pltp) term
  and appp term = (sep_end_by1 appbp spaces1 >>= fun ts -> return (app ts)) term
	and appbp term = (pltp <|> letp <|> ifp <|> consp <|> absp <|> varp) term
	and varp term = ((varname) >>= fun v -> return (var v)) term
	and letp term = (tempt (string "let " >> spaces >> varname >>= fun v -> spaces >> string "=" >> spaces >> ltup (string "in " >> spaces) >>= fun t -> ltp >>= fun b -> return (leti v t b))) term
	and ifp term = (tempt (string "if " >> spaces >> ltup (string "then " >> spaces) >>= fun c -> ltup (string "else " >> spaces) >>= fun t1 -> ltp >>= fun t2 -> return (ite c t1 t2))) term
	and consp term = ((tempt (string "true" >> return (cons True))) <|> (tempt (string "false" >> return (cons False)))) term
	and ltup until term = (many_until (ltup' until) until >>= fun ts -> return (if Lst.length ts = 1 then Lst.hd ts else app ts)) term
  and ltup' until term = ((pltp <|> letp <|> ifp <|> consp <|> absup until <|> varp) >>= fun t -> spaces >> return t) term
	and absup until term = (char '\\' >> abshp >>= fun t1 -> char '.' >> absbup until >>= fun t2 -> return (abs t1 t2)) term
	and absbup until term = (many_before (ltup' until) until >>= fun ts -> return (if Lst.length ts = 1 then Lst.hd ts else app ts)) term;;
		
(* PRINTER-FUNCTIONS *) 
(* some helper-functions *)
let var2strng (id, i) =
 if i <= 0 then id else Lst.append id (Strng.of_int i);;
let rec get_nested_abs_vars = function 
	| Abs (v, l) -> v :: get_nested_abs_vars l
	| _ -> [];;
let rec get_nested_abs_term = function 
	| Abs (v, l) -> get_nested_abs_term l
	| l -> l;;
let rec vars2strng = function 
	| [] -> []
	| x::[] -> (var2strng x)
	| x::xs -> (var2strng x) @ (' '::(vars2strng xs));;
let cons2strng = function 
	| True -> Strng.of_string "true"
	| False -> Strng.of_string "false";;

(** printer-function for CoreML-terms, without conventions used *)
let rec to_strng' = function
	| Var v -> var2strng v
	| Abs (v, t) ->
			'(' :: '\\' :: Lst.append (Lst.append (var2strng v) ('.' :: to_strng' t)) [')']
	| App (t1, t2) ->
			'(' :: Lst.append (Lst.append (to_strng' t1) (' ' :: to_strng' t2)) [')']
	| Let (v, t, b) -> 
		  ['(';'l';'e';'t';' '] @ (var2strng v) @ ['='] @ (to_strng' t) @ [' ';'i';'n';' '] @ (to_strng' b) @ [')']
	| If (c, t1, t2) -> 
			['i';'f';' '] @ (to_strng' c) @ [' ';'t';'h';'e';'n';' '] @ (to_strng' t1) @ [' ';'e';'l';'s';'e';' '] @ (to_strng' t2)
	| Cons c -> cons2strng c;;

let rec to_strng = function 
	| Var v -> var2strng v
	| Abs (v, t) -> let vs = v::get_nested_abs_vars t 
									and ls = get_nested_abs_term t in 
									'\\' :: (Lst.append (Lst.append (vars2strng vs) ('.' :: (to_strng ls))) [])
	| App (Abs (v1, t1), Abs (v2, t2)) -> '(' :: Lst.append (Lst.append (to_strng (Abs (v1, t1))) [')';' ']) ('(' :: Lst.append (to_strng (Abs (v2, t2))) [')'])
	| App (Abs (v1, t1), App(t2, t3)) -> '(' :: Lst.append (to_strng (Abs (v1, t1))) (')' :: ' ' :: '(' :: Lst.append (to_strng (App (t2, t3))) [')'])
	| App (Abs (v1, t1), t2) -> '(' :: Lst.append (Lst.append (to_strng (Abs (v1, t1))) [')';' ']) (to_strng t2)
	| App (t1, Abs (v2, t2)) -> Lst.append (to_strng t1) (' ' :: '(' :: Lst.append (to_strng (Abs (v2, t2))) [')'])
	| App (t1, App(t2, t3)) -> Lst.append (to_strng t1) (' ' :: '(' :: Lst.append (to_strng (App (t2, t3))) [')'])
	| App (t1, t2) -> Lst.append (to_strng t1) (' ' :: to_strng t2)
	| Let (v,t, b) -> ['l';'e';'t';' '] @ (var2strng v) @ ['='] @ (to_strng t) @ [' ';'i';'n';' '] @ (to_strng b)
	| If (c, t1, t2) -> ['i';'f';' '] @ (to_strng c) @ [' ';'t';'h';'e';'n';' '] @ (to_strng t1) @ [' ';'e';'l';'s';'e';' '] @ (to_strng t2)
	| Cons c -> cons2strng c;;

let to_strng_mark ps t = 
	let mark_on = Strng.of_string "\027[04m" in
	let mark_off = Strng.of_string "\027[00m" in
	let rec to_strng pos t = if Lst.mem pos ps then mark_on @ to_strng' pos t @ mark_off else to_strng' pos t 
	and to_strng' pos = function
		| Var v -> var2strng v
		| Abs (v, t) -> let vs = v::get_nested_abs_vars t 
										and ls = get_nested_abs_term t in 
										'\\' :: (Lst.append (Lst.append (vars2strng vs) ('.' :: (to_strng (pos@(Lst.init Left (Lst.length vs))) ls))) [])
		| App (Abs (v1, t1), Abs (v2, t2)) -> '(' :: Lst.append (Lst.append (to_strng (pos@[Left]) (Abs (v1, t1))) [')';' ']) ('(' :: Lst.append (to_strng (pos@[Right]) (Abs (v2, t2))) [')'])
		| App (Abs (v1, t1), App(t2, t3)) -> '(' :: Lst.append (to_strng (pos@[Left]) (Abs (v1, t1))) (')' :: ' ' :: '(' :: Lst.append (to_strng (pos@[Right]) (App (t2, t3))) [')'])
		| App (Abs (v1, t1), t2) -> '(' :: Lst.append (Lst.append (to_strng (pos@[Left]) (Abs (v1, t1))) [')';' ']) (to_strng (pos@[Right]) t2)
		| App (t1, Abs (v2, t2)) -> Lst.append (to_strng (pos@[Left]) t1) (' ' :: '(' :: Lst.append (to_strng (pos@[Right]) (Abs (v2, t2))) [')'])
		| App (t1, App(t2, t3)) -> (*'(' ::*) Lst.append (to_strng (pos@[Left]) t1) (' ' :: '(' :: Lst.append (to_strng (pos@[Right]) (App (t2, t3))) [')'])
		| App (t1, t2) -> Lst.append (to_strng (pos@[Left]) t1) (' ' :: to_strng (pos@[Right]) t2)
		| Let (v,t, b) -> ['l';'e';'t';' '] @ (var2strng v) @ ['='] @ (to_strng (pos@[Left]) t) @ [' ';'i';'n';' '] @ (to_strng (pos@[Right]) b)
		| If (c, t1, t2) -> ['i';'f';' '] @ (to_strng (pos@[Left]) c) @ [' ';'t';'h';'e';'n';' '] @ (to_strng (pos@[Middle]) t1) @ [' ';'e';'l';'s';'e';' '] @ (to_strng (pos@[Right]) t2)
		| Cons c -> cons2strng c
	in to_strng [] t;;

let to_strng_mark' ps t = 
	let mark_on = Strng.of_string "\027[04m" in
	let mark_off = Strng.of_string "\027[00m" in
	let rec to_strng pos t = if Lst.mem pos ps then mark_on @ to_strng' pos t @ mark_off else to_strng' pos t 
	and to_strng' pos = function
		| Var v -> var2strng v
		| Abs (v, t) -> ['(';'\\'] @ (var2strng v) @ ['.'] @ (to_strng' (pos@[Left]) t) @ [')']
		| App (t1, t2) -> ['('] @ (to_strng (pos@[Left]) t1) @ [' '] @ (to_strng (pos@[Right]) t2) @ [')']
		| Let (v,t, b) -> ['(';'l';'e';'t';' '] @ (var2strng v) @ ['='] @ (to_strng (pos@[Left]) t) @ [' ';'i';'n';' '] @ (to_strng (pos@[Right]) b) @ [')']
		| If (c, t1, t2) -> ['(';'i';'f';' '] @ (to_strng (pos@[Left]) c) @ [' ';'t';'h';'e';'n';' '] @ (to_strng (pos@[Middle]) t1) @ [' ';'e';'l';'s';'e';' '] @ (to_strng (pos@[Right]) t2) @ [')']
		| Cons c -> cons2strng c
	in to_strng [] t;;



