(* Copyright 2008 Christian Sternagel *)

(* This file is part of LIPS.

  LIPS is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as 
	published by the Free Software Foundation, either version 3 of 
	the License, or (at your option) any later version.

  LIPS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
	License along with LIPS.  If not, see 
	<http://www.gnu.org/licenses/>. *)

type 'a reply    = Error | Ok of ('a * Strng.t);;
type 'a consumed = Empty of 'a | Consumed of 'a;;
type 'a result   = ('a reply) consumed;;
type 'a t = Strng.t -> 'a result;;

let reply = function
 | Empty r -> r
 | Consumed r -> r
;;

let parse p s = match reply (p s) with
 | Error -> None
 | Ok (x, _) -> Some x
;;

let test p s = match parse p (Strng.of_string s) with 
	| None -> failwith "Parser.test: no successful parse possible"
	| Some x -> x;;

let sat p = function
 | [] -> Empty Error
 | c :: cs -> if p c then Consumed (Ok (c, cs)) else Empty Error
;;

let any_char = sat (fun _ -> true);;
let char c = sat ((=) c);;
let letter = sat (fun c -> (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'));;
let digit  = sat (fun c -> c >= '0' && c <= '9');;

let (>>=) p f = (fun s -> match p s with
 | Consumed r -> begin match r with
  | Ok (x, s') -> begin match f x s' with
   | Empty r' -> Consumed r'
   | consumed -> consumed
  end
  | Error -> Consumed Error
 end
 | Empty r -> begin match r with
  | Ok (x, s') -> f x s'
  | Error -> Empty Error
 end
);;

let (>>) p q = p >>= fun _ -> q;;
let return x = fun s -> Empty (Ok (x, s));;

let (<|>) p q = (fun s -> match p s with
 | Empty Error -> q s
 | other -> other
);;

let tempt p = (fun s -> match p s with
 | Consumed Error -> Empty Error
 | other -> other
);;

let rec many_accum accum p = (fun s ->
 let rec walk xs s r = match r with
  | Empty Error -> Ok (xs, s)
  | Empty ok    -> Error
  | Consumed Error -> Error
  | Consumed (Ok (x, s')) -> let ys = accum x xs in walk ys s' (p s')
 in match p s with
  | Empty r -> begin match r with
   | Ok _ -> Empty Error
   | Error -> Empty (Ok ([], s))
  end
  | consumed -> Consumed (walk [] s consumed)
);;

let many p =
 many_accum (fun x xs -> x :: xs) p >>= fun xs -> return (Lst.rev xs)
;;

let many1 p = p >>= fun x -> many p >>= fun xs -> return (x :: xs);;

let sep_by1 p sep =
 p >>= fun x -> many (sep >> p) >>= fun xs -> return (x :: xs)
;;

let sep_by p sep = sep_by1 p sep <|> return [];;

let rec sep_end_by1 p sep =
 p >>= fun x -> (
  (sep >> sep_end_by p sep >>= fun xs -> return (x :: xs))
  <|> return [x]
 )
and sep_end_by p sep = sep_end_by1 p sep <|> return [];;

let between o p c = o >> p >>= fun x -> c >> return x;;
let skip_many p = many_accum (fun x xs -> []) p >> return ();;
let skip_many1 p = p >> skip_many p;;
let one_of s = sat (fun c -> String.contains s c);;
let none_of s = sat (fun c -> not (String.contains s c));;
let space = one_of " \n\t";;
let spaces = skip_many space;;
let spaces1 = space >> spaces;;

let string s = (fun input ->
 let s = Strng.of_string s in
 let rec walk = function
  | ([], cs) -> Ok (s, cs)
  | (xs, []) -> Error
  | ((x :: xs), (c :: cs)) -> if x = c then walk (xs, cs)
                                       else Error
 and walk1 = function
  | ([], cs) -> Empty (Ok (s, cs))
  | (xs, []) -> Empty Error
  | ((x :: xs), (c :: cs)) -> if x = c then Consumed (walk (xs, cs))
                                       else Empty Error 
 in walk1 (s, input)
);;



let many_before_accum accum p u = (fun s -> 
	let rec walk xs s = match u s with 
		| Empty Error -> begin match p s with 
			| Empty Error -> Error
			| Empty ok -> Error
			| Consumed Error -> Error
			| Consumed (Ok (x, s')) -> let ys = accum x xs in walk ys s' end
		| Empty ok -> Error
		| Consumed Error -> begin match p s with 
			| Empty Error -> Error
			| Empty ok -> Error
			| Consumed Error -> Error
			| Consumed (Ok (x, s')) -> let ys = accum x xs in walk ys s' end
		| Consumed (Ok (_, _)) -> (Ok (xs, s))
  in Consumed (walk [] s)
);;
let many_before p b = 
	many_before_accum (fun x xs -> x::xs) p b >>= fun xs -> return (Lst.rev xs);;
let many_until p u =  
	many_before_accum (fun x xs -> x::xs) p u >>= fun xs -> u >> return (Lst.rev xs);;



let parse_strict p s = match reply (p s) with
 | Ok (x, []) -> Some x
 | _ -> None
;;

