(* Copyright 2008, 2009 Michael Brunner *)

(* This file is part of LIPS.

  LIPS is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as 
	published by the Free Software Foundation, either version 3 of 
	the License, or (at your option) any later version.

  LIPS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
	License along with LIPS.  If not, see 
	<http://www.gnu.org/licenses/>. *)

type t = Var of int | Arrow of t * t | TCons of Strng.t * (t list);;
type env = (CoreML.t * t) St.t;;

let var i = Var i;;
let arrow t1 t2 = Arrow (t1, t2);; 
let tcons s ts = TCons (s, ts);;
let tcons' s ts = tcons (Strng.of_string s) ts;;

(* DEFINITIONS FOR BASIC ENVIRONMENT *)
(* basic types *)
let bool = tcons' "bool" [];;
let int = tcons' "int" [];;
(* basic environment *)
let basic_env = St.of_list [
	(CoreML.cons CoreML.True, bool);
	(CoreML.cons CoreML.False, bool);
	];;


(* FUNCTIONS FOR TYPE-CHECKING AND TYPE_INFERENCE *)
(** retrieve type-variables for given type *)
let rec tvar = function 
	| Arrow (t1, t2) -> St.union (tvar t1) (tvar t2)
	| TCons (s, t::ts) -> (Lst.fold (fun x y -> St.insert x y) (St.singleton t) ts)
	| t -> St.singleton t;;
(** retrieve maximum variable-index occuring within a given type *)
let max_index v = St.fold (fun x y -> match x with Var i -> max i y | _ -> y) 0 (St.filter (function Var _ -> true | _ -> false) (tvar v));;

(** retrieve index for next free variable in environment *)
let next_free_index env = 
	let vars = St.fold (fun x y -> St.insert (snd x) y) St.empty env in
	let i = St.fold (fun x y -> max (max_index x) y) 0 vars
	in (i + 1);;

(** increment all variable-indexes within a given type *)
let rec inc_var_index i = function 
	| Var n -> Var (n + i)
	| Arrow (t1, t2) -> Arrow (inc_var_index i t1, inc_var_index i t2)
	| TCons (s, ts) -> TCons (s, Lst.map (inc_var_index i) ts);;

(** calculate unification-problem for term with type t under environment env *)
let get_unification env term t = 
	let rec gu acc env' term t = 
		let i = max (next_free_index acc) ((max_index t) + 1) in
		if St.has_one (fun x -> fst x = term) env' then
			St.insert (t, snd (St.find_by (fun x -> fst x = term) env')) acc
		else if St.has_one (fun x -> fst x = term) env then
			St.insert (t, inc_var_index i (snd (St.find_by (fun x -> fst x = term) env))) acc
		else 
			match term with 
				| CoreML.Var _ -> acc
				| CoreML.Cons _ -> acc
				| CoreML.Abs (x, e) -> gu (St.insert (t, arrow (var i) (var (i + 1))) acc) (St.insert (CoreML.Var x, (var i)) env') e (var (i + 1))
				| CoreML.App (e1, e2) -> gu (gu acc env' e1 (arrow (var i) t)) env' e2 (var i)
				| CoreML.Let (x, e1, e2) -> gu (gu acc env' e1 (var i)) (St.insert (CoreML.Var x, var i) env') e2 t
				| CoreML.If (c, e1, e2) -> gu (gu (gu acc env' c bool) env' e1 t) env' e2 t  
	in gu St.empty St.empty term t;;

(** apply type-substitution to a single type *)
let rec tsub sub = function 
	| Arrow (t1, t2) -> arrow (tsub sub t1) (tsub sub t2)
	| TCons(s, t::ts) -> tcons s (Lst.map (tsub sub) (t::ts))
	| t -> Lst.fold (fun (x, x') y -> match x with Var _ when x = y -> x' | _ -> y) t sub;;

(** reduce var-indexes to minimum *)
let reduce_var_index t = 
	let rec init i j = if j >= i then [var i] else (var j)::(init i (j + 1)) in
	let vars = Lst.sort (fun x y -> match (x,y) with (Var i, Var j) -> i - j | _ -> 0) (St.to_list (St.filter (fun x -> match x with Var _ -> true | _ -> false) (tvar t))) in
	let length = Lst.length vars in 
	if length > 0 then 
		let nvars = init (Lst.length vars) 0 in
		tsub (Lst.rev (Lst.zip2 vars nvars)) t
	else 
		t;;
	 
(** apply rules of type-inference on a single type-equation *)
let apply_inference_rules = function 
		| TCons (s1, ts), TCons (s2, ts') when s1 = s2 -> (St.of_list (Lst.zip2 ts ts'), [])
		| Arrow (t1, t2), Arrow (t1', t2') -> (St.of_list ([(t1, t1'); (t2, t2')]), [])
		| Var a, t when not (St.mem (Var a) (tvar t)) -> (St.singleton (Var a, t), [(Var a, t)])
		| t, Var a when not (St.mem (Var a) (tvar t)) -> (St.singleton (Var a, t), [(Var a, t)])
		| t, t' when t = t' -> (St.empty, [])
		| r -> (St.singleton r, []);;

(** check wether the given unification-problem is satisfiable *)
let is_unifiable problem = 
	let folder rule problem = 
		let nrule, subst = apply_inference_rules rule in
		let nprob = St.union nrule (St.remove rule problem) in
	  if subst = [] then 
			nprob
		else 
			St.fold (fun (x, x') nprob' -> St.insert (tsub subst x, tsub subst x') nprob') St.empty nprob in 
	let rec reduce problem = 
		let nprob = St.fold folder problem problem in
		if problem = nprob then 
			problem 
		else 
			reduce nprob in
	St.is_empty (reduce problem);;
	
(** check if type t can be derived for term under environment env *)
let check_type env term t = 
	is_unifiable (get_unification env term t);;

(** check if the given term is typable under environment env *)
let is_typable env term = 
	is_unifiable (get_unification env term (var 0));;

(** infer the most general type of the given term under environment env *)
let infer_type env term = 
	let up = get_unification env term (var 0) in
	let folder rule (problem, substitutions) = 
		let nrule, nsubst = apply_inference_rules rule in 
		let nprob = St.union nrule (St.remove rule problem) in 
		let nprob' = if nsubst = [] then nprob else St.fold (fun (x, x') nprob' -> St.insert (tsub nsubst x, tsub nsubst x') nprob') St.empty nprob in 
		let nsubst' = (Lst.map (fun (t, t') -> (t, tsub nsubst t')) substitutions) @ nsubst in
		(nprob', nsubst') in
	let rec reduce (problem, substitution) = 
		let (nprob, nsubst) = St.fold folder (problem, substitution) problem in
		if problem = nprob then 
			(problem, substitution)
		else 
			reduce (nprob, nsubst) in
	match reduce (up, []) with 
		| (prob, substitutions) when St.is_empty prob -> reduce_var_index (tsub substitutions (var 0))
		| _ -> failwith "Type.infer_type: no type could be infered";;
	
	

(* PARSERS *)
open Parser;; (* needed for infix-operators of module Parser *)

(** parser for type-constructor-identifiers *)
let conname term = 
	(letter >>= fun p1 -> many (char '_' <|> letter <|> digit) >>= fun p2 -> return (p1::p2)) term;;

(** parser for types *)
let rec typep term = (arrowp <|> varp <|> consp <|> ptypep) term
  and varp term = (char '\'' >> many1 (sat (fun c -> c >= 'a' && c <= 'z')) >>= fun ls -> return (var (Lst.fold1 (fun x y -> x * 10 + y) (Lst.map (fun x -> if x >= 'a' && x <= 'z' then (Char.code x) - (Char.code 'a') else 0) ls)))) term
	and arrowp term = (tempt ((ptypep <|> varp <|> consp) >>= fun t1 -> spaces >> string "->" >> spaces >> typep >>= fun t2 -> return (arrow t1 t2))) term
	and consp term = (tempt (conname >>= fun s -> spaces >> consp' >>= fun ts -> return (tcons s ts))) term
	and consp' term = ((between (spaces >> char '(' >> spaces) (sep_by typep (spaces >> char ',' >> spaces)) (spaces >> char ')' >> spaces) >>= fun ts -> return ts) <|> (return [])) term
	and ptypep term = (between (char '(' >> spaces) typep (spaces >> char ')')) term;;



(* PRINTER-FUNCTIONS*)
(** return lstring for type-variable *)
let var_to_strng i = 
	let rec index_to_chars i = 
		if i < 26 then 
			[Char.chr (i + (Char.code 'a'))]
		else 
			(Char.chr ((i mod 26) + (Char.code 'a')))::(index_to_chars (i / 26))
	in '\''::(index_to_chars i);;
(** return lstring of given type *)
let rec to_strng = function 
	| Var i -> var_to_strng i
	| Arrow (Arrow (t1, t1'), t2) -> ['('] @ (to_strng t1) @ [' ';'-';'>';' '] @ (to_strng t1') @ [')'] @ [' ';'-';'>';' '] @ (to_strng t2)
	| Arrow (t1, t2) -> (to_strng t1) @ [' ';'-';'>';' '] @ (to_strng t2)
	| TCons (s, []) -> s
	| TCons (s, ts) -> s @ ['('] @ (Lst.fold1 (fun x y -> x @ [',';' '] @ y) (Lst.map to_strng ts)) @ [')'];;

(** return string-representation of gven type *)
let to_string t = Strng.to_string (to_strng t);;
	
(** top-level-printing-function *)
let toplevel_printer fmt s = Format.fprintf fmt "\"%s\"" (String.escaped (Strng.to_string (to_strng s)));;
